# Generated by roxygen2: do not edit by hand

export(savvy_glm.fit2)
export(savvy_glm2)
importFrom(CVXR,Minimize)
importFrom(CVXR,Problem)
importFrom(CVXR,Variable)
importFrom(CVXR,psolve)
importFrom(CVXR,sum_squares)
importFrom(CVXR,value)
importFrom(MASS,ginv)
importFrom(Matrix,Schur)
importFrom(Matrix,rankMatrix)
importFrom(expm,"%^%")
importFrom(glm2,glm.fit2)
importFrom(glm2,glm2)
importFrom(glmnet,cv.glmnet)
importFrom(parallel,detectCores)
importFrom(parallel,makeCluster)
importFrom(parallel,mclapply)
importFrom(parallel,parLapply)
importFrom(parallel,stopCluster)
importFrom(stats,.getXlevels)
importFrom(stats,binomial)
importFrom(stats,coef)
importFrom(stats,gaussian)
importFrom(stats,is.empty.model)
importFrom(stats,lm)
importFrom(stats,model.extract)
importFrom(stats,model.frame)
importFrom(stats,model.matrix)
importFrom(stats,model.offset)
importFrom(stats,model.response)
importFrom(stats,model.weights)
importFrom(stats,poisson)
importFrom(stats,predict)
importFrom(utils,globalVariables)
